//////////////////////////////////
//	Sohail Hirani				//
//	sah77@pitt.edu				//
//	Assignment 6 CS1501			//
//	Graph implementations		//
//////////////////////////////////




#include <fstream.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>


struct edge{
	//elements in edgeQueue
	int v1;
	int v2;
	int weight;
};

struct list {
	int value;
	int priority;
	struct list *next;
};

class queue {

	// A priority queue as needed by the algorithm
	// For non operation priority the priority is just set to 0
private:
	struct list *start;

public:
	queue() {
		start = NULL;
	}

	void insert(int value) {
		// Non priority insert
		insert(value, 0);
	}

	void insert(int value,int priority) {
		//Insert with a certain priority
		if( start == NULL ) {
			// The list is empty
			start = new struct list[1];
			start->value = value;
			start->priority = priority;
			start->next = NULL;
		}else{
			struct list *node = start;
			struct list *prev = NULL;
			while( node != NULL  && node->priority <= priority) {
				// Find the right place for the element
				// Convention followed is place it at the end of elements
				// with same priority but before those with greater priority value

				prev = node;
				node = node->next;
			}
			if( prev == NULL ) {
				// This means the value should be at the begining
				prev = start;
				start = new struct list[1];
				start->value = value;
				start->priority = priority;
				start->next = prev;
			}else{
				// Put value at the appropriate location
				prev->next = new struct list[1];
				prev = prev->next ;
				prev->value = value;
				prev->priority = priority;
				prev->next = node;
			}
		}
	}

	int update(int value,int priority) {
		// If element value exists in the queue 
		// and if its priority is greater than the passed priority 
		// change priority and return 1;
		// If value not present insert into queue
		// otherwise return 0 to say no change occured to the queue
		if( start == NULL ) {
			// Insert at the begining
			start = new struct list[1];
			start->value = value;
			start->priority = priority;
			start->next = NULL;
			return 1;
		}else{
			struct list *node = start;
			struct list *prev = NULL;
			while( node != NULL && node->value != value) {
				// Search for the value
				prev = node;
				node = node->next;
			}
			if( node == NULL ) {
				// Value not found in queue
				insert(value,priority);
				return 1;
			}else{
				if( node->priority > priority ) {
					if( prev == NULL) {
						node->priority = priority;
					}else { 
						// Delete the node and add it again with the new priority
						struct list *temp = node;
						prev->next = node->next;
						delete[] node;
						insert(value,priority);
					}
					return 1;
				}
			}
		}
		return 0;
	}

	int remove(void) {
		// Return the first element in the queue
		if( start == NULL ) {
			return -1;
		}
		int value = start->value;
		struct list *temp = start;
		start = start->next;
		delete [] temp;
		return value;
	}

	void clear() {
		// Empty the queue
		while(start)
			remove();
	}

	int isEmpty() {
		if( start == NULL )
			return 1;
		else 
			return 0;
	}

};

struct edgeList {
	struct edge value;
	struct edgeList *next;
};

class edgeQueue {
//Useful to store the edges inorder while doing BFS and DFS 
//  cause we have to print the edges after the connected components
private:
	struct edgeList *start;
	struct edgeList *end; //Using end speeds up insertions

public:
	edgeQueue() {
		start = NULL;
		end = NULL;
	}

	void insert(struct edge value) {
		//Insert the edge to the queue
		if( start == NULL ) {
			start = new struct edgeList[1];
			end = start;
			start->value = value;
			end->next = NULL;
		}else{
			end->next = new struct edgeList[1];
			end = end->next;
			end->value = value;
			end->next = NULL;
		}
	}

	struct edge remove(void) {
		//Remove the first edge
		if( start == NULL ) {
			struct edge dummy;
			dummy.v1 = -1;
			dummy.v2 = -1;
			return dummy;
		}
		struct edge value = start->value;
		struct edgeList *temp = start;
		start = start->next;
		delete [] temp;
		if( start == NULL )
			end = NULL;
		return value;
	}

	void clear() {
		// Empty the Queue
		while(start)
			remove();
	}

	int isEmpty() {
		if( start == NULL )
			return 1;
		else 
			return 0;
	}

};

class Graph {

private:
	int **adjacencyMatrix;
	int v;	// Number of vertices
	int e;	// Number of edges
	bool isWeighted;	// True indicates weighted graph 
	char **vertices;

	void makeBFSTree(	int vertex,
					int *seen,
					queue & cc,
					edgeQueue & edges
				  ) {
		// This build a BFS tree from a give node
		// It is supposed to be used only by doBFS hence private

		queue vertexQueue;
		vertexQueue.insert(vertex);
		int curVertex;
		int i;
		while( (curVertex = vertexQueue.remove())!= -1 ) {
			for(i=0; i<v; i++) {
				if( adjacencyMatrix[curVertex][i] > 0 && !seen[i] ) {
					seen[i] = 1;
					struct edge temp;
					temp.v1 = curVertex;
					temp.v2 = i;
					temp.weight = adjacencyMatrix[curVertex][i];
					edges.insert(temp);
					cc.insert(i);
					vertexQueue.insert(i);
				}
			}
		}
	}

	void makeDFSTree(	int vertex,
					int *seen,
					queue & cc,
					edgeQueue & edges
				  ) {
		// Makes a DFS tree from give node/"vertex"

		int i;
		for(i=0; i<v; i++) {
			if( adjacencyMatrix[vertex][i] > 0 && !seen[i] ) {
				seen[i] = 1;
				struct edge temp;
				temp.v1 = vertex;
				temp.v2 = i;
				temp.weight = adjacencyMatrix[vertex][i];
				edges.insert(temp);
				cc.insert(i);
				makeDFSTree(i,seen,cc,edges);
			}
		}
	}


public:
	Graph() {
		adjacencyMatrix = NULL;
		v = 0;
		e = 0;
		isWeighted = false;
		vertices = NULL;
	}

	Graph(const Graph& g) {
		// Copy constructor
		int i;
		int j;
		v = g.v;
		e = g.e;
		isWeighted = g.isWeighted;
		adjacencyMatrix = (int **)malloc( sizeof(int *) * v);
		for( i=0; i<v; i++ ) {
			adjacencyMatrix[i] = new int[v];
			for( j=0; j<v; j++ ) {
				// Copy each value
				adjacencyMatrix[i][j] = g.adjacencyMatrix[i][j];
			}
		}
		vertices = (char **)malloc( sizeof(char *) * v);
		for( i=0; i<v; i++ ) {
			// Asigne space that is just enough
			vertices[i] = new char[strlen(g.vertices[i])+1];
			strcpy(vertices[i],g.vertices[i]);
		}
	}

	~Graph() {
		// Destructor - just clear the memory allocated by new/malloc
		int i;
		if( adjacencyMatrix != NULL ) {
			for(i=0; i<v; i++) {
				delete [] adjacencyMatrix[i];
			}
			free(adjacencyMatrix);
		}
		if( vertices != NULL ) {
			for(i=0; i<v; i++) {
				if( vertices[i] != NULL )
					delete[] vertices[i];
			}
			free(vertices);
		}
	}

	Graph operator = (const Graph & g) {
		// Overload the assignment operator
		int i;
		int j;
		if( adjacencyMatrix != NULL ) {
			for(i=0; i<v; i++) {
				delete [] adjacencyMatrix[i];
			}
			free(adjacencyMatrix);
		}
		if( vertices != NULL ) {
			for(i=0; i<v; i++) {
				if( vertices[i] != NULL )
					delete[] vertices[i];
			}
			free(vertices);
		}
		v = g.v;
		e = g.e;
		isWeighted = g.isWeighted;
		adjacencyMatrix = (int **)malloc( sizeof(int *) * v);
		for( i=0; i<v; i++ ) {
			adjacencyMatrix[i] = new int[v];
			for( j=0; j<v; j++ ) {
				adjacencyMatrix[i][j] = g.adjacencyMatrix[i][j];
			}
		}
		vertices = (char **)malloc( sizeof(char *) * v);
		for( i=0; i<v; i++ ) {
			vertices[i] = NULL;
			setVertex(i,g.vertices[i]);
		}
		return *this;
	}

	void setWeighted(bool isWeighted) {
		this->isWeighted = isWeighted;
	}

	void setVertices(int v) {
		// Initialization part where vertice is given 
		// now we have to allocate memory locations accordingly
		int i;
		int j;
		//Clear previous allocations
		if( adjacencyMatrix != NULL ) {
			for(i=0; i<v; i++) {
				delete [] adjacencyMatrix[i];
			}
			free(adjacencyMatrix);
		}
		if( vertices != NULL ) {
			for(i=0; i<v; i++) {
				if( vertices[i] != NULL )
 					delete[] vertices[i];
			}
			free(vertices);
		}
		this->v = v;
		// New allocation
		adjacencyMatrix = (int **)malloc(sizeof(int *) * v);
		for(i=0; i<v; i++ ) {
			adjacencyMatrix[i] = new int[v];
		}

		vertices = (char **)malloc(sizeof(char *) * v);
		for(i=0; i<v; i++) {
			vertices[i] = NULL;
		}
		
		for(i=0; i<v; i++) {
			for(j=0; j<v; j++) {
				adjacencyMatrix[i][j] = 0;
			}
		}
	}

	void setEdges(int e) {
		this->e = e;
	}

	void setEdge(int e1, int e2, int weight) {
		// Since its a square matrix
		adjacencyMatrix[e1-1][e2-1] = weight;
		adjacencyMatrix[e2-1][e1-1] = weight;
	}

	void setVertex(int vertex, char *vertexName) {
		if(vertices[vertex] != NULL)
			free(vertices[vertex]);
		vertices[vertex] = new char[strlen(vertexName)+1];
		strcpy(vertices[vertex],vertexName);
	}

	void parseFile(char *filename) {
		// Give the filename now parse this file and get your data
		int i;
		int e1;
		int e2;
		int weight;
		bool isWeighted = false;
		char buf[256];
		fstream fin(filename, ios::in);
		if( !fin.is_open() ) {
			cerr << "Error : could not open file " << filename;
			return;
		}
		fin>>i;
		if( i == 1)
			isWeighted = true;
		setWeighted(isWeighted);
		fin>>v;
		setVertices(v);
		fin>>e;
		setEdges(e);
		for(i=0; i<e; i++ ){
			// Get the edges
			weight = 1;
			fin>>e1;
			fin>>e2;
			if( isWeighted ) 
				fin>>weight;
			setEdge(e1,e2,weight);
		}
		for(i=0; i<v; i++) {
			// Read the names from the for 
			fin>>buf;
			setVertex(i,buf);
		}
		return;
	}

	void display() {
		// Display the value for the graph
		int i;
		int j;
		if( isWeighted )
			cout << "\nType of Graph : Weighted";
		else
			cout << "\nType of Graph : Unweighted";
		cout << "\nNumber of vertices " << v ;
		cout << "\nNumber of edges " << e ;
		for(i=0; i<v; i++) {
			for(j=i+1; j<v; j++) {
				if( adjacencyMatrix[i][j] > 0 ) {
					cout << "\n" << vertices[i] << "<->" << vertices[j] ;
					if( isWeighted ) {
						cout << ", weight = " << adjacencyMatrix[i][j];
					}
				}
			}
		}
	}


	

	void doBFS() {

		// Perform Breath first search on the graph
		int i;
		int ccCnt = 0;
		int *seen = new int[v];
		for(i=0; i<v; i++)
			seen[i] = 0;

		queue cc;
		edgeQueue edges;
		int vertex;

		cout<< "\nDoing BFS on the graph";
		for(i=0; i<v; i++) {
			if( !seen[i] ) {
				seen[i] = 1;
				cc.insert(i);
				makeBFSTree(i,seen,cc,edges);
				// Build BFS Tree for this node
			}
			if( !cc.isEmpty() ) {
				cout<<"\nConnected Component " << ++ccCnt <<" : ";
				// Print the vertices
				cout << "\nVertices are :";
			}

			while( (vertex=cc.remove())!=-1 ) {
				cout << vertex+1 << " ";
				cout.flush();
			}
			// Print the edges
			if( !edges.isEmpty() )
				cout << "\nEdges are : ";
			struct edge temp;
			while( !edges.isEmpty()  ) {
				temp = edges.remove();
				cout << "\n" << vertices[temp.v1] << "<->"
					 << vertices[temp.v2] ;
				if( isWeighted ) 
					cout << ", weight = " << temp.weight;
				cout.flush();
			}
		}

		delete[] seen;
	}




	void doDFS() {

		//Perform Depth first search on the graph
		int i;
		int ccCnt = 0;
		int *seen = new int[v];
		for(i=0; i<v; i++)
			seen[i] = 0;

		queue cc;
		edgeQueue edges;
		int vertex;

		cout<< "\nDoing DFS on the graph";
		for(i=0; i<v; i++) {
			if( !seen[i] ) {
				seen[i] = 1;
				cc.insert(i);
				makeDFSTree(i,seen,cc,edges);
				//Build DFS tree for the vertex
			}
			if( !cc.isEmpty() ) {
				cout<<"\nConnected Component " << ++ccCnt <<" : ";
				cout << "\nVertices are :";
			}
			while( (vertex=cc.remove())!=-1 ) {
				cout << vertex+1 << " ";
				cout.flush();
			}
			if( !edges.isEmpty() )
				cout << "\nEdges are : ";
			struct edge temp;
			while( !edges.isEmpty()  ) {
				temp = edges.remove();
				cout << "\n" << vertices[temp.v1] << "<->"
					 << vertices[temp.v2] ;
				if( isWeighted ) 
					cout << ", weight = " << temp.weight;
				// if weighted print the weight
				cout.flush();
			}
		}

		delete[] seen;
	}

	void findSPWeighted(int v1, int v2) {

		// Find the shortest path for a weighted graph
		// use prims algorithm in this case

		int *seen = new int[v];
		int *priority = new int[v];
		int *parent = new int[v];
		int i;
		queue pq;
		queue vertexQueue;
		int curVertex;

		v1--;
		v2--;
		for(i=0; i<v; i++)
			seen[i] = 0;
		for(i=0; i<v; i++)
			priority[i] = 0;
		for(i=0; i<v; i++)
			parent[i] = 0;
		parent[v2] = -1;
		pq.insert(v2);
		
		while( !pq.isEmpty() ) {
			curVertex = pq.remove();
			if(curVertex == v1) {
				pq.clear();
				break;
			}
			seen[curVertex] = 1;
			for( i=0; i<v; i++ ) {
				if( !seen[i] && adjacencyMatrix[curVertex][i] > 0 ) {
					if( pq.update(i,adjacencyMatrix[curVertex][i] + priority[curVertex]) ) {
						priority[i] = adjacencyMatrix[curVertex][i] + priority[curVertex];
						parent[i] = curVertex;
					}
				}
			}
		}

		if( curVertex != v1 ) {
			cout << "\nNot path linking " << vertices[v1] 
				 << ", " << vertices[v2] << " found";
		}else{
			cout << "\nShortest Path found : total weight " << priority[v1];
			while( parent[curVertex]!=-1 ) {
					cout << "\n" << vertices[curVertex]
						 << "<->" << vertices[parent[curVertex]]
						 << ", weight = " 
						 << adjacencyMatrix[curVertex][parent[curVertex]];
					curVertex = parent[curVertex];
			}
		}
	}

	void findSPUnweighted(int v1, int v2) {

		// Find shortes path for a unweighted graph
		// Use BFS for this case
		int *seen = new int[v];
		int *priority = new int[v];
		int *parent = new int[v];
		int i;
		queue pq;
		int curVertex;

		v1--;
		v2--;
		for(i=0; i<v; i++)
			seen[i] = 0;
		for(i=0; i<v; i++)
			priority[i] = 0;
		for(i=0; i<v; i++)
			parent[i] = 0;
		parent[v2] = -1;
		seen[v2] = 1;
		pq.insert(v2);
		
		while( !pq.isEmpty() ) {
			curVertex = pq.remove();
			if(curVertex == v1) {
				pq.clear();
				break;
			}
			for( i=0; i<v; i++ ) {
				if( !seen[i] && adjacencyMatrix[curVertex][i] > 0 ) {
					seen[i] = 1;
					pq.insert(i);
					priority[i] = 1 + priority[curVertex];
					parent[i] = curVertex;
				}
			}
		}

		if( curVertex != v1 ) {
			cout << "\nNot path linking " << vertices[v1] 
				 << ", " << vertices[v2] << " found";
		}else{
			cout << "\nShortest Path found : edges " << priority[v1];
			while( parent[curVertex]!=-1 ) {
					cout << "\n" << vertices[curVertex]
						 << "<->" << vertices[parent[curVertex]]
						 << ", weight = " 
						 << adjacencyMatrix[curVertex][parent[curVertex]];
					curVertex = parent[curVertex];
			}
		}
	}


	void findShortestPath(int v1, int v2) {
		if( isWeighted )
			findSPWeighted(v1,v2);
		else
			findSPUnweighted(v1,v2);
	}

};


void main(int argc, char *argv[]) {

	char filename[30];
	if(argc == 1) {
		cout<<"\nEnter filename : ";
		cin>>filename;
	}else strcpy(filename, argv[1]);

	Graph g;
	g.parseFile(filename);
	cout << "Initial graph";
	g.display();
	cout << "\nTesting the copy constructor";
	Graph h(g);	//Test the copy constructor
	cout << "\nResult of copy constructor";
	Graph k;
	cout << "\nTesting the assignment operator";
	k = g;
	cout << "\nResult of assigned operator";
	h.display();

	// Perform BFS
	g.doBFS();
	//Prform DFS
	g.doDFS();

	int v1;
	int v2;
	cout << "\nEnter v1 v2 for shortest path: ";
	cin>>v1>>v2;
	g.findShortestPath(v1,v2); 
	// The funtion check for appropriate
	cout<<"\n";
}
